<?php

declare(strict_types=1);

namespace app\models;

use yii;

use app\models\traits\SearchByEdge;

use moonland\phpexcel\Excel;

/**
 * Продукт (в ассортименте магазина)
 *
 * Представляет собой лот состоящий из предложений от поставщиков
 *
 * @see Supply Поставки для продуктов
 */
class Product extends Document
{
    use SearchByEdge;

    /**
     * Сценарий импорта из .excel документа
     *
     * Использовать для обхода правил при загрузке файла
     */
    const SCENARIO_IMPORT = 'import';

    /**
     * Сценарий записи товара
     */
    const SCENARIO_WRITE = 'write';

    /**
     * Файл .excel для импорта товаров
     */
    public Excel|string|array|null $file = null;

    /**
     * Группа в которой состоит товар
     */
    public ProductGroup|null $group = null;

    /**
     * Имя коллекции
     */
    public static function collectionName(): string
    {
        return 'product';
    }

    /**
     * Свойства
     */
    public function attributes(): array
    {
        return array_merge(
            parent::attributes(),
            [
                'desc',
                'ocid',
                'catn',
                'imgs',
                'time',
                'oemn',
                'cost'
            ]
        );
    }

    /**
     * Метки свойств
     */
    public function attributeLabels(): array
    {
        return array_merge(
            parent::attributeLabels(),
            [
                'catn' => 'Каталожный номер (catn)',
                'desc' => 'Описание (desc)',
                'ocid' => 'Идентификатор 1C (ocid)',
                'imgs' => 'Изображения (imgs)',
                'time' => 'Срок доставки (time)',
                'oemn' => 'OEM номера (oemn)',
                'cost' => 'Стоимость (cost)',
                'file' => 'Документ (file)',
                'group' => 'Группа (group)'
            ]
        );
    }

    /**
     * Правила
     */
    public function rules(): array
    {
        return array_merge(
            parent::rules(),
            [
                [
                    'catn',
                    'required',
                    'message' => 'Заполните поля: {attribute}',
                    'on' => self::SCENARIO_WRITE,
                    'except' => self::SCENARIO_IMPORT
                ],
                [
                    'catn',
                    'string',
                    'message' => '{attribute} должен быть строкой'
                ],
                [
                    [
                        'oemn',
                        'imgs'
                    ],
                    'arrayValidator',
                    'message' => '{attribute} должен быть массивом.'
                ],
                [
                    'file',
                    'required',
                    'message' => 'Заполните поля: {attribute}',
                    'on' => self::SCENARIO_IMPORT
                ],
                [
                    'file',
                    'file',
                    'skipOnEmpty' => false,
                    'extensions' => 'xlsx',
                    'checkExtensionByMimeType' => false,
                    'maxFiles' => 5,
                    'maxSize' => 1024 * 1024 * 30,
                    'wrongExtension' => 'Разрешены только документы в формате: ".xlsx"',
                    'message' => 'Проблема при чтении документа',
                    'on' =>  self::SCENARIO_IMPORT
                ]
            ]
        );
    }

    /**
     * Инициализация продукта
     *
     * @param string $catn Артикул, каталожный номер
     */
    public static function initEmpty(string $catn): self|array
    {
        $oemn = self::searchOemn($catn);

        if (count($oemn) === 1) {
            // Передан только один артикул

            if ($model = self::searchByCatn($catn)) {
                // Продукт уже существует

                return $model;
            }

            // Запись пустого продукта
            return self::writeEmpty($catn);
        }

        // Инициализация
        $models = [];

        foreach ($oemn as $catn) {
            // Перебор всех найденных артикулов

            if ($model = self::searchByCatn($catn)) {
                // Продукт уже существует

                continue;
            }

            // Запись
            if ($model = self::writeEmpty($catn)) {
                // Записано

                // Запись в массив сохранённых моделей
                $models[] = $model;
            }
        }

        return $models;
    }

    /**
     * Запись пустого продукта
     */
    public static function writeEmpty(string $catn): ?self
    {
        // Инициализация
        $model = new self;

        // Настройки
        $model->catn = $catn;

        // Запись
        return $model->save() ? $model : null;
    }

    /**
     * Поиск OEM номеров
     *
     * @param string $oemn Необработанная строка с OEM-номерами
     * @param string $delimiters Разделители
     *
     * @todo НЕ ЗАБЫТЬ СДЕЛАТЬ НАСТРОЙКУ РАЗДЕЛИТЕЛЕЙ
     *
     * @return array OEM-номера
     */
    public static function searchOemn(string $oemn, string $delimiters = '\s\+\/,'): array
    {
        // Инициализация
        $catn = [];

        // Конвертация
        preg_match_all("/[^$delimiters]+/", $oemn, $catn);

        return $catn[0];
    }

    /**
     * Импорт товаров
     *
     * На данный момент обрабатывает только импорт из
     * файлов с расширением .excel
     */
    public function import(): bool
    {
        // Инициализация
        $data = [];
        $amount = 0;

        if ($this->validate()) {
            foreach ($this->file as $file) {
                // Перебор файлов

                // Инициализация
                $dir = '../assets/import/' . date('Y_m_d#H-i', time()) . '/excel/';

                // Сохранение на диск
                if (!file_exists($dir)) {
                    mkdir($dir, 0775, true);
                }
                $file->saveAs($path = $dir . $file->baseName . '.' . $file->extension);

                $data[] = Excel::import($path, [
                    'setFirstRecordAsKeys' => true,
                    'setIndexSheetByName' => true,
                ]);
            }


            foreach ($data as $data) {
                // Перебор конвертированных файлов

                if (count($data) < 1) {
                    // Не найдены строки с товарами

                    $this->addError('erros', 'Не удалось найти данные товаров');
                } else {
                    // Перебор найденных товаров

                    foreach ($data as $doc) {
                        // Перебор полученных документов

                        // Сохранение в базе данных
                        $product = new static($doc);

                        $product->scenario = $product::SCENARIO_WRITE;

                        if ($product->validate()) {
                            // Проверка пройдена

                            // Запись документа
                            $product->save();

                            // Постинкрементация счётчика
                            $amount++;

                            // Запись группы
                            // $group = static::class . 'Group';
                            // (new $group())->writeMember($product, $this->group);
                        } else {
                            // Проверка не пройдена
                            foreach ($product->errors as $attribute => $error) {
                                $this->addError($attribute, $error);
                            }
                        }
                    }
                }
            }

            // Деинициализация
            $this->file = '';

            static::afterImportExcel($amount);

            return true;
        }

        $this->addError('erros', 'Неизвестная ошибка');

        static::afterImportExcel($amount);

        return false;
    }

    /**
     * Поиск по каталожному номеру
     *
     * Ищет продукт и возвращает его,
     * либо выполняет поиск через представление
     *
     * @todo Переделать нормально
     */
    public static function searchByCatn(string $catn, int $limit = 1, array $select = []): static|array|null
    {
        if ($limit <= 1) {
            return static::findOne(['catn' => $catn]);
        }

        $query = self::find()
            ->collection('product_search')
            ->search(['catn' => $catn])
            ->limit($limit)
            ->select($select)
            ->createCommand()
            ->execute()
            ->getAll();

        foreach ($query as &$attribute) {
            // Приведение всех свойств в массив и очистка от лишних данных

            $attribute = $attribute->getAll();
        }

        return $query;
    }

    /**
     * Вызывается после загрузки поставок из excel-документа
     *
     * @param int $amount Количество
     */
    public static function afterImportExcel(int $amount = 0): bool
    {
        // Инициализация
        $model = new Notification;
        $date = date('H:i d.m.Y', time());

        // Настройка
        $model->text = yii::$app->controller->renderPartial('/notification/system/afterImportExcel', compact('amount', 'date'));
        $model->type = $model::TYPE_NOTICE;

        // Отправка
        return (bool) $model->write();
    }

    /**
     * Вызывается после загрузки поставок из 1С
     *
     * @param int $amount Количество
     */
    public static function afterImportOnec(): bool
    {
        // Инициализация
        $model = new Notification;
        $date = date('H:i d.m.Y', time());

        // Настройка
        $model->text = yii::$app->controller->renderPartial('/notification/system/afterImportOnec', compact('amount', 'date'));
        $model->type = $model::TYPE_NOTICE;

        // Отправка
        return (bool) $model->write();
    }
}
